/*
 * Real-time clock support.
 *
 * Copyright (C) 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "rtc-defs.h"
#include "rtc-generic.h"



// Peripheral abstraction.

Rtc_chip::Rtc_chip(l4_addr_t addr, Cpm_chip *cpm)
: _cpm(cpm)
{
  _regs = new Hw::Mmio_register_block<32>(addr);

  // Reset time regulation, in case it is completely scrambled.
  // NOTE: Using 32768 cycles for a 1Hz signal.

  set_regulator(32767, 0);
}

enum Clock_identifiers
Rtc_chip::get_clock()
{
  return Clock_rtc;
}

uint32_t
Rtc_chip::read_checked(unsigned reg)
{
  uint32_t last, current;

  wait();
  last = _regs[reg];

  while (1)
  {
    wait();
    current = _regs[reg];

    if (current == last)
      return current;
    else
      last = current;
  }
}

void
Rtc_chip::wait()
{
  while (!(_regs[Rtc_control] & Control_write_ready));
}

void
Rtc_chip::write_enable()
{
  wait();
  _regs[Hibernate_write_enable_pattern] = Write_enable_pattern;

  while (!(_regs[Hibernate_write_enable_pattern] & Write_enable_status));

  wait();
}

void
Rtc_chip::disable()
{
  write_enable();
  _regs[Rtc_control] = _regs[Rtc_control] & ~Control_rtc_enable;
}

void
Rtc_chip::enable()
{
  write_enable();
  _regs[Rtc_control] = _regs[Rtc_control] | Control_rtc_enable;
}

void
Rtc_chip::alarm_disable()
{
  write_enable();
  _regs[Rtc_control] = _regs[Rtc_control] & ~Control_alarm_enable;
}

void
Rtc_chip::alarm_enable()
{
  write_enable();
  _regs[Rtc_control] = (_regs[Rtc_control] & ~Control_alarm) | Control_alarm_enable;
}

void
Rtc_chip::wakeup_alarm_disable()
{
  write_enable();
  _regs[Hibernate_wakeup_control] = _regs[Hibernate_wakeup_control] & ~Rtc_alarm_wakeup_enable;
}

void
Rtc_chip::wakeup_alarm_enable()
{
  write_enable();
  _regs[Hibernate_wakeup_control] = _regs[Hibernate_wakeup_control] | Rtc_alarm_wakeup_enable;
}

uint32_t
Rtc_chip::get_seconds()
{
  return read_checked(Rtc_seconds);
}

void
Rtc_chip::set_seconds(uint32_t seconds)
{
  write_enable();
  _regs[Rtc_seconds] = seconds;
}

uint32_t
Rtc_chip::get_alarm_seconds()
{
  return read_checked(Rtc_alarm_seconds);
}

void
Rtc_chip::set_alarm_seconds(uint32_t seconds)
{
  write_enable();
  _regs[Rtc_alarm_seconds] = seconds;
}

void
Rtc_chip::set_regulator(uint32_t base, uint32_t adjustment)
{
  base = base ? base - 1 : 0;
  adjustment = adjustment ? adjustment - 1 : 0;

  if (base > Regulator_1Hz_cycle_count_limit)
    base = Regulator_1Hz_cycle_count_limit;

  if (adjustment > Regulator_adjust_count_limit)
    adjustment = Regulator_adjust_count_limit;

  write_enable();
  _regs[Rtc_regulator] = (base << Regulator_1Hz_cycle_count_shift) |
                         (adjustment << Regulator_adjust_count_shift);
}

// Device-specific method.

void
Rtc_chip::_pre_power_down()
{
}

void
Rtc_chip::_power_down()
{
  _pre_power_down();

  write_enable();
  _regs[Hibernate_control] = _regs[Hibernate_control] | Hibernate_power_down;
}

void
Rtc_chip::hibernate()
{
  alarm_enable();
  wakeup_alarm_enable();
  _power_down();
}

void
Rtc_chip::power_down()
{
  wakeup_alarm_disable();
  _power_down();
}
